
/******************************************************************************

Copyright (c) 1997, Andrew J. Weber, Jason P. Weber

Copyright information should have been included with this distribution
and may be found at ../meta/COPYRIGHT or obtained via email from weber@acm.org

******************************************************************************/

#include <bld/varop.h>
#include <bld/interp.h>

/***********************************************************************
class BLD_SetOperation
***********************************************************************/
BLD_Value BLD_SetOperation::Perform(void) const
{
	BLD_Value value;

	// If the previous value should not be preserved...
	if (BLD_SET_OVERWRITE == type)
	{
		value = op->Perform();
	}
	else
	{
		// This is inefficient, but it is a simple way of
		// ensuring that no null values are inserted in
		// values.
		const BLD_Value previousValue = BLD_CONTEXT.GetValue(name);
		register const unsigned long n = previousValue.GetCount();
		for(register unsigned long i = 0; n > i; ++i)
		{
			if (!previousValue[i].IsNull())
			{
				value.AppendElement(previousValue[i]);
			}
		}

		// If the previous value goes to the end...
		if (BLD_SET_PREPEND == type)
		{
			value.InsertRange(op->Perform(), 0);
		}

		// If the previous value stays at the front...
		else if (BLD_SET_APPEND == type)
		{
			value.AppendRange(op->Perform());
		}
	}

	BLD_CONTEXT.SetValue(name, value);
	return value;
}

GPL_String BLD_SetOperation::ToString(void) const
{
	GPL_String image;
	switch(type)
	{
	case BLD_SET_PREPEND:
		image.Printf("set %s -= %s", name.GetData(), op->ToString().GetData());
		break;

	case BLD_SET_OVERWRITE:
		image.Printf("set %s = %s", name.GetData(), op->ToString().GetData());
		break;

	case BLD_SET_APPEND:
		image.Printf("set %s += %s", name.GetData(), op->ToString().GetData());
		break;

	default:
		break;
	}

	return image;
}

/***********************************************************************
class BLD_GetOperation
***********************************************************************/
BLD_Value BLD_GetOperation::Perform(void) const
{
	const BLD_Value value = BLD_CONTEXT.GetValue(name);
	return value;
}

GPL_String BLD_GetOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("$(%s)", name.GetData());
	return image;
}

/***********************************************************************
class BLD_BeginScopeOperation
***********************************************************************/
BLD_Value BLD_BeginScopeOperation::Perform(void) const
{
	const BLD_Scope scope(name);
	if (BLD_CONTEXT.Push(scope))
	{
		return scope.GetName();
	}
	else
	{
		return BLD_Value::falsehood;
	}
}

GPL_String BLD_BeginScopeOperation::ToString(void) const
{
	GPL_String image;
	image.Printf("begin %s", name.GetData());
	return image;
}

/***********************************************************************
class BLD_EndScopeOperation
***********************************************************************/
BLD_Value BLD_EndScopeOperation::Perform(void) const
{
	BLD_Scope scope;
	if (BLD_CONTEXT.Pop(scope))
	{
		return scope.GetName();
	}
	else
	{
		return BLD_Value::falsehood;
	}
}

GPL_String BLD_EndScopeOperation::ToString(void) const
{
	return "end";
}
